# encoding: UTF-8
#
# Copyright 2016 Sam Oluwalana <soluwalana@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

require 'date'
require 'logging'
require 'httpclient'
require 'builder'
require 'gyoku'
require 'base64'
require 'nori'
require 'rexml/document'
require 'securerandom'
require 'thread'
require 'cgi'

require_relative 'version'
require_relative 'transport'

require_relative 'wsmv/commands/init_runspace_pool'
require_relative 'wsmv/commands/create_pipeline'
require_relative 'wsmv/commands/close_shell'
require_relative 'wsmv/commands/send_data'
require_relative 'wsmv/command_output_processor'


module PSRP

  class PSRPService

    DEFAULT_TIMEOUT = 60
    DEFAULT_MAX_ENV_SIZE = 153600
    DEFAULT_LOCALE = 'en-US'

    attr_accessor :xfer, :session_opts

    def initialize(endpoint, opts = {})
      @session_opts = {
        endpoint: endpoint,
        max_envelope_size: DEFAULT_MAX_ENV_SIZE,
        session_id: SecureRandom.uuid.to_s.upcase,
        operation_timeout: DEFAULT_TIMEOUT,
        locale: DEFAULT_LOCALE
      }

      @logger = Logging.logger[self]
      @logger.level = opts[:log_level] || :debug
      @logger.add_appenders(Logging.appenders.stdout)
      @opts = opts

      @xfer = HTTP::Negotiate.new(endpoint, @opts[:user], @opts[:pass], @opts)

      @shell_id = nil
      @generated_shell_id = nil
      @opened = false
    end

    def open
      @logger.debug("[WinRM] opening remote runspacepool on #{@session_opts[:endpoint]}")
      msg = PSRP::WSMV::InitRunspacePool.new(@session_opts)

      resp_doc = @xfer.send_request(msg.build)

      @generated_shell_id = msg.shell_id
      @shell_id = REXML::XPath.first(resp_doc, "//*[@Name='ShellId']").text
      

      out_processor = PSRP::WSMV::CommandOutputProcessor.new(@session_opts, @xfer)
      
      out_processor.command_output(@shell_id, nil)
      return false if out_processor.command_done?

      while true
        out_processor.command_output(@shell_id, nil, true)
        if out_processor.command_done?
          break
        end

        runspace_open = false
        
        # check to make sure the runspace is opened
        out_processor.defragmented.each do |msg_id, msg|
          if msg[:message_type] == PSRP::Message::MESSAGE_TYPES[:RUNSPACEPOOL_STATE]
            xml = REXML::Document.new(msg[:data])
            REXML::XPath.match(xml, "//*[@N='RunspaceState']").each do |n|
              runspace_open = (n.text == '2')
            end
          end
        end
        
        if runspace_open
          @opened = true
          @logger.debug("[WinRM] remote runspace #{@shell_id} is open on #{@session_opts[:endpoint]}")
          return true
        end
      end

      return false
    end

    def run_ps(command, shell_opts = {})

      # spec says:
      # The client sends a wxf:Receive message (section 3.1.5.3.7) to the server to start receiving data from 
      # the server. After each received wxf:ReceiveResponse message (section 3.2.5.3.8), the client sends another 
      # wxf:Receive message until the RunspacePool transitions to a Closed or Broken state.
      if not @opened
        open
      end

      out_processor = PSRP::WSMV::CommandOutputProcessor.new(@session_opts, @xfer)
      
      @logger.debug('Runspace is open - sending command')
            
      command_id = SecureRandom.uuid.to_s.upcase
      pipeline = PSRP::MessageEncoder.new(@generated_shell_id, command_id, :CREATE_PIPELINE, {command: CGI.escapeHTML(command)})

      msg = PSRP::WSMV::CreatePipeline.new(@session_opts, @shell_id, pipeline.fragments[0])
      resp_doc = @xfer.send_request(msg.build)

      command_id = REXML::XPath.first(resp_doc, "//#{PSRP::WSMV::NS_WIN_SHELL}:CommandId").text

      # Send remaining fragments, if any
      if pipeline.fragments.length > 1
        for i in 1..(pipeline.fragments.length - 1)
          msg = PSRP::WSMV::SendData.new(@session_opts, @shell_id, command_id, pipeline.fragments[i])
          resp_doc = @xfer.send_request(msg.build)
        end
      end

      @logger.debug("Command Sent, waiting on responses")
      out_processor.command_output(@shell_id, command_id, true)
      if out_processor.input_required?
        raise PSRPError.new('This Tool Does Not Support Accepting Runtime Input')
      end
      
      # retrieve all of the data from the command
      while true  
        out_processor.command_output(@shell_id, command_id)
        if out_processor.input_required?
          raise PSRPError.new('This Tool Does Not Support Accepting Runtime Input')
        end
        if out_processor.command_done?
          break
        end
      end 

      # Deal with fragmented return values
      datas = out_processor.defragmented

      # Deal With Errors
      # But not really doing it well
      # PSRP - 2.2.3.15 ErrorRecord
      if out_processor.has_error?
        datas.each do |msg_id, msg|
          if msg[:message_type] == PSRP::Message::MESSAGE_TYPES[:ERROR_RECORD]
            raise PSRPError.new(un_psrp_escape(msg[:data]))
          end
        end
      end

      data = ''
      out_processor.defragmented.each do |msg_id, msg|
        xml = REXML::Document.new(msg[:data])
        # TODO: do better deserialization based on MS-PSRP 2.2.5 
        # Right now this only returns String primitives and String Extended Primitives
        REXML::XPath.match(xml, "/S|/Obj/S").each do |n|
          next if n.text.nil? || n.text.empty?
          data += n.text + "\n"
        end
      end
      return data
    end

    def close
      if not @opened
        return
      end
      @opened = false
      @logger.debug("Closing shell")
      msg = PSRP::WSMV::CloseShell.new(@session_opts, @shell_id)
      resp_doc = @xfer.send_request(msg.build)
    end    

    def un_psrp_escape(str)
      begin
        str.gsub(/_x([0-9A-F]{4})_/i) { [Regexp.last_match[1].hex].pack('U') }
      rescue Encoding::CompatibilityError
        str
      end
    end
  end
end

