# encoding: UTF-8
#
# Copyright 2016 Shawn Neal <sneal@sneal.net>
# Copyright 2016 Sam Oluwalana <soluwalana@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 

require_relative "response_handler"

require 'monitor'

module PSRP
  module HTTP

    # straight copy from WinRM
    # NTLM/Negotiate, secure, HTTP transport
    class Negotiate
      
      DEFAULT_RECEIVE_TIMEOUT = 1000
      
      def initialize(endpoint, user, pass, opts)
        @endpoint = endpoint.is_a?(String) ? URI.parse(endpoint) : endpoint
        @httpcli = HTTPClient.new(agent_name: 'Ruby PSRP Client')
        @httpcli.receive_timeout = DEFAULT_RECEIVE_TIMEOUT
        @httpcli.connect_timeout = 2
        @httpcli.send_timeout = 2
        @logger = Logging.logger[self]
        require 'rubyntlm'

        # Deleting SSPI authentication
        auths = @httpcli.www_auth.instance_variable_get('@authenticator')
        auths.delete_if { |i| i.is_a? HTTPClient::SSPINegotiateAuth }

        user_parts = user.split('\\')
        if(user_parts.length > 1)
          opts[:domain] = user_parts[0]
          user = user_parts[1]
        end

        @ntlmcli = Net::NTLM::Client.new(user, pass, opts)
        @retryable = true
        no_ssl_peer_verification! if opts[:no_ssl_peer_verification]
        @ssl_peer_fingerprint = opts[:ssl_peer_fingerprint]
        @httpcli.ssl_config.set_trust_ca(opts[:ca_trust_path]) if opts[:ca_trust_path]

      end

      # Disable SSL Peer Verification
      def no_ssl_peer_verification!
        @httpcli.ssl_config.verify_mode = OpenSSL::SSL::VERIFY_NONE
      end

      # SSL Peer Fingerprint Verification prior to connecting
      def ssl_peer_fingerprint_verification!
        return unless @ssl_peer_fingerprint && ! @ssl_peer_fingerprint_verified

        with_untrusted_ssl_connection do |connection|
          connection_cert = connection.peer_cert_chain.last
          verify_ssl_fingerprint(connection_cert)
        end
        @logger.info("initial ssl fingerprint #{@ssl_peer_fingerprint} verified\n")
        @ssl_peer_fingerprint_verified = true
        no_ssl_peer_verification!
      end

      # Connect without verification to retrieve untrusted ssl context
      def with_untrusted_ssl_connection
        noverify_peer_context = OpenSSL::SSL::SSLContext.new
        noverify_peer_context.verify_mode = OpenSSL::SSL::VERIFY_NONE
        tcp_connection = TCPSocket.new(@endpoint.host, @endpoint.port)
        begin
          ssl_connection = OpenSSL::SSL::SSLSocket.new(tcp_connection, noverify_peer_context)
          ssl_connection.connect
          yield ssl_connection
        ensure
          tcp_connection.close
        end
      end

      # compare @ssl_peer_fingerprint to current ssl context
      def verify_ssl_fingerprint(cert)
        return unless @ssl_peer_fingerprint
        conn_fingerprint = OpenSSL::Digest::SHA1.new(cert.to_der).to_s
        return unless @ssl_peer_fingerprint.casecmp(conn_fingerprint) != 0
        fail "ssl fingerprint mismatch!!!!\n"
      end

      # HTTP Client receive timeout. How long should a remote call wait for a
      # for a response from WinRM?
      def receive_timeout=(sec)
        @httpcli.receive_timeout = sec
      end

      def receive_timeout
        @httpcli.receive_timeout
      end


      def send_request(message, auth_header = nil)
        ssl_peer_fingerprint_verification!
        auth_header = init_auth if @ntlmcli.session.nil?

        original_length = message.length

        emessage = @ntlmcli.session.seal_message message
        signature = @ntlmcli.session.sign_message message
        seal = "\x10\x00\x00\x00#{signature}#{emessage}"
        
        hdr = {
          "Content-Type" => "multipart/encrypted;protocol=\"application/HTTP-SPNEGO-session-encrypted\";boundary=\"Encrypted Boundary\""
        }
        hdr.merge!(auth_header) if auth_header

        # p hdr

        body = [
          "--Encrypted Boundary",
          "Content-Type: application/HTTP-SPNEGO-session-encrypted",
          "OriginalContent: type=application/soap+xml;charset=UTF-8;Length=#{original_length}",
          "--Encrypted Boundary",
          "Content-Type: application/octet-stream",
          "#{seal}--Encrypted Boundary--",
          ""
        ].join("\r\n")

        @logger.debug("\nOut-Message\n\n")
        doc = REXML::Document.new message
        out = ""
        doc.write(out, 2)
        @logger.debug(out)
        @logger.debug("\n\n")

        resp = @httpcli.post(@endpoint, body, hdr)
        verify_ssl_fingerprint(resp.peer_cert)
        if resp.status == 401 && @retryable
          @retryable = false
          send_request(message, init_auth)
        else
          
                    
          @retryable = true
          decrypted_body = resp.body.empty? ? '' : decrypt(resp.body)
          handler = PSRP::ResponseHandler.new(decrypted_body, resp.status)
          data = handler.parse_to_xml()
          
          @logger.debug("Response data\n\n")
          doc = REXML::Document.new decrypted_body
          out = ""
          doc.write(out, 2)
          @logger.debug(out)
          
          data
        end
      end

      private

      def decrypt(str)
        str.force_encoding('BINARY')
        str.sub!(/^.*Content-Type: application\/octet-stream\r\n(.*)--Encrypted.*$/m, '\1')

        signature = str[4..19]
        message = @ntlmcli.session.unseal_message str[20..-1]
        if @ntlmcli.session.verify_signature(signature, message)
          message
        else
          raise PSRP::PSRPError, "Could not verify SOAP message."
        end
      end

      def init_auth
        auth1 = @ntlmcli.init_context
        hdr = {"Authorization" => "Negotiate #{auth1.encode64}",
               "Content-Type" => "application/soap+xml;charset=UTF-8"
        }
        @logger.debug("Sending HTTP POST for Negotiate Authentication")
        r = @httpcli.post(@endpoint, "", hdr)
        verify_ssl_fingerprint(r.peer_cert)
        itok = r.header["WWW-Authenticate"].pop.split.last
        binding = r.peer_cert.nil? ? nil : Net::NTLM::ChannelBinding.create(r.peer_cert)
        auth3 = @ntlmcli.init_context(itok, binding)
        { "Authorization" => "Negotiate #{auth3.encode64}" }
      end

      def log_soap_message(message)
        return unless @logger.debug?
        xml_msg = REXML::Document.new(message)
        formatter = REXML::Formatters::Pretty.new(2)
        formatter.compact = true
        formatter.write(xml_msg, @logger)
        @logger.debug("\n")
      rescue StandardError => e
        @logger.debug("Couldn't log SOAP request/response: #{e.message} - #{message}")
      end
    end
  end
end

