# encoding: UTF-8
#
# Copyright 2016 Shawn Neal <sneal@sneal.net>
# Copyright 2016 Sam Oluwalana <soluwalana@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

require_relative 'commands/receive'

module PSRP

  module WSMV
    # Class to handle getting all the output of a command until it completes
    class CommandOutputProcessor
      
      attr :msgs

      INPUT_REQUIRED = [
        PSRP::Message::MESSAGE_TYPES[:PIPELINE_HOST_CALL],
        PSRP::Message::MESSAGE_TYPES[:RUNSPACEPOOL_HOST_CALL]
      ]

      # Creates a new command output processor
      # @param connection_opts [ConnectionOpts] The WinRM connection options
      # @param transport [HttpTransport] The WinRM SOAP transport
      # @param out_opts [Hash] Additional output options
      def initialize(connection_opts, transport, out_opts = {})
        @connection_opts = connection_opts
        @transport = transport
        @out_opts = out_opts
        @has_error = false
        @command_done = false
        @input_required = false
        @msgs = {}
      end

      # Gets the command output from the remote shell
      # @param shell_id [UUID] The remote shell id running the command
      # @param command_id [UUID] The command id to get output for
      # @param block Optional callback for any output
      def command_output(shell_id, command_id, reset = false)
        if reset == true
          @has_error = false
          @command_done = false
          @input_required = false
        end

        out_message = command_output_message(shell_id, command_id)
        resp_doc = send_get_output_message(out_message)
        streams(resp_doc)
        @command_done = REXML::XPath.match(
          resp_doc,
          "//*[@State='http://schemas.microsoft.com/wbem/wsman/1/windows/shell/" \
          "CommandState/Done']").any?
        resp_doc
      end

      def command_output_message(shell_id, command_id)
        cmd_out_opts = {
          shell_id: shell_id
        }.merge(@out_opts)
        if command_id
          cmd_out_opts[:command_id] = command_id
        end
        PSRP::WSMV::ReceiveOutput.new(@connection_opts, cmd_out_opts).build
      end

      def send_get_output_message(message)
        resp_doc = @transport.send_request(message)
      rescue PSRP::WSManFault => e
        # If no output is available before the wsman:OperationTimeout expires,
        # the server MUST return a WSManFault with the Code attribute equal to
        # 2150858793. When the client receives this fault, it SHOULD issue
        # another Receive request.
        # http://msdn.microsoft.com/en-us/library/cc251676.aspx
        if e.fault_code == '2150858793'
          retry
        else
          raise
        end
      end

      def exit_code(resp_doc)
        REXML::XPath.first(resp_doc, "//#{NS_WIN_SHELL}:ExitCode").text.to_i
      end

      def streams(resp_doc)
        REXML::XPath.match(resp_doc, "//#{PSRP::WSMV::NS_WIN_SHELL}:Stream").each do |n|
          next if n.text.nil? || n.text.empty?
          msg = decode(n.text)
          if not @msgs.has_key? msg.message_id
            @msgs[msg.message_id] = []
          end
          @msgs[msg.message_id].push(msg)
          if INPUT_REQUIRED.include? msg.message_type
            @input_required = true
          elsif msg.message_type == PSRP::Message::MESSAGE_TYPES[:ERROR_RECORD]
            @has_error = true
          end
        end
      end

      # Defragment Messages
      def defragmented
        datas = {}
        @msgs.each do |msg_id, msg_list|
          msg_list.sort do |a, b|
            a.fragment_id <=> b.fragment_id
          end
          unfragmented = msg_list.inject('') do |data, msg|
            data += msg.data
          end
          datas[msg_id] = {
            message_type: msg_list[0].message_type,
            data: unfragmented,
          }
        end
        datas
      end

      def has_error?
        @has_error
      end

      def input_required?
        @input_required
      end

      def command_done?
        @command_done
      end

      # Decode the raw PSRP output into decoded and human consumable object
      
      # @param raw_output [String] The raw encoded output
      # @return [String] The decoded output
      def decode(raw_output)
        # TODO: Add better decoding based on MS-PSRP 2.2.5
        PSRP::MessageDecoder.new(raw_output)
      end
    end
  end
end
