# encoding: UTF-8
#
# Copyright 2016 Shawn Neal <sneal@sneal.net>
# Copyright 2016 Sam Oluwalana <soluwalana@gmail.com>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

require_relative '../psrp_message'

module Iso8601Duration
  # Convert the number of seconds to an ISO8601 duration format
  # @see http://tools.ietf.org/html/rfc2445#section-4.3.6
  # @param [Fixnum] seconds The amount of seconds for this duration
  def self.sec_to_dur(seconds)
    seconds = seconds.to_i
    iso_str = 'P'
    if seconds > 604_800 # more than a week
      weeks = seconds / 604_800
      seconds -= (604_800 * weeks)
      iso_str << "#{weeks}W"
    end
    if seconds > 86_400 # more than a day
      days = seconds / 86_400
      seconds -= (86_400 * days)
      iso_str << "#{days}D"
    end
    if seconds > 0
      iso_str << 'T'
      if seconds > 3600 # more than an hour
        hours = seconds / 3600
        seconds -= (3600 * hours)
        iso_str << "#{hours}H"
      end
      if seconds > 60 # more than a minute
        minutes = seconds / 60
        seconds -= (60 * minutes)
        iso_str << "#{minutes}M"
      end
      iso_str << "#{seconds}S"
    end

    iso_str
  end
end

module PSRP
  module WSMV
    # Base class for all WSMV message classes

    NS_SOAP_ENV    = 's'   # http://www.w3.org/2003/05/soap-envelope
    NS_ADDRESSING  = 'a'   # http://schemas.xmlsoap.org/ws/2004/08/addressing
    NS_CIMBINDING  = 'b'   # http://schemas.dmtf.org/wbem/wsman/1/cimbinding.xsd
    NS_ENUM        = 'n'   # http://schemas.xmlsoap.org/ws/2004/09/enumeration
    NS_TRANSFER    = 'x'   # http://schemas.xmlsoap.org/ws/2004/09/transfer
    NS_WSMAN_DMTF  = 'w'   # http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd
    NS_WSMAN_MSFT  = 'p'   # http://schemas.microsoft.com/wbem/wsman/1/wsman.xsd
    NS_SCHEMA_INST = 'xsi' # http://www.w3.org/2001/XMLSchema-instance
    NS_WIN_SHELL   = 'rsp' # http://schemas.microsoft.com/wbem/wsman/1/windows/shell
    NS_WSMAN_FAULT = 'f'   # http://schemas.microsoft.com/wbem/wsman/1/wsmanfault
    NS_WSMAN_CONF  = 'cfg' # http://schemas.microsoft.com/wbem/wsman/1/config

    # WSMan URI of the regular Windows cmd shell
    RESOURCE_URI_CMD = 'http://schemas.microsoft.com/wbem/wsman/1/windows/shell/cmd'
      
    # WSMan URI for PowerShell
    RESOURCE_URI_POWERSHELL = 'http://schemas.microsoft.com/powershell/Microsoft.PowerShell'

    class Base

      # Builds the WSMV message XML payload
      def build
        builder = Builder::XmlMarkup.new
        builder.instruct!(:xml, encoding: 'UTF-8')
        builder.tag! :env, :Envelope, namespaces do |env|
          env.tag!(:env, :Header) do |env_header|
            create_header(env_header)
          end
          env.tag!(:env, :Body) do |env_body|
            create_body(env_body)
          end
        end
      end

      def namespaces
        @namespaces ||= {
          'xmlns:xsd' => 'http://www.w3.org/2001/XMLSchema',
          'xmlns:xsi' => 'http://www.w3.org/2001/XMLSchema-instance',
          'xmlns:env' => 'http://www.w3.org/2003/05/soap-envelope',
          "xmlns:#{NS_ADDRESSING}" => 'http://schemas.xmlsoap.org/ws/2004/08/addressing',
          "xmlns:#{NS_CIMBINDING}" => 'http://schemas.dmtf.org/wbem/wsman/1/cimbinding.xsd',
          "xmlns:#{NS_ENUM}" => 'http://schemas.xmlsoap.org/ws/2004/09/enumeration',
          "xmlns:#{NS_TRANSFER}" => 'http://schemas.xmlsoap.org/ws/2004/09/transfer',
          "xmlns:#{NS_WSMAN_DMTF}" => 'http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd',
          "xmlns:#{NS_WSMAN_MSFT}" => 'http://schemas.microsoft.com/wbem/wsman/1/wsman.xsd',
          "xmlns:#{NS_WIN_SHELL}" => 'http://schemas.microsoft.com/wbem/wsman/1/windows/shell',
          "xmlns:#{NS_WSMAN_CONF}" => 'http://schemas.microsoft.com/wbem/wsman/1/config'
        }
      end

      protected

      def create_header
        fail 'create_header must be implemented'
      end

      def create_body
        fail 'create_body must be implemented'
      end

      def encode_bytes(bytes)
        Base64.strict_encode64(bytes.pack('C*'))
      end

      # Merge the various header hashes and make sure we carry all of the attributes
      # through instead of overwriting them.
      def merge_headers(*headers)
        hdr = {}
        headers.each do |h|
          hdr.merge!(h) do |k, v1, v2|
            v1.merge!(v2) if k == :attributes!
          end
        end
        hdr
      end

      def shared_headers(session_opts)
        {
          "#{NS_ADDRESSING}:To" => "#{session_opts[:endpoint]}",
          "#{NS_ADDRESSING}:ReplyTo" => {
            "#{NS_ADDRESSING}:Address" =>
              'http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous',
            :attributes! => {
              "#{NS_ADDRESSING}:Address" => {
                'mustUnderstand' => true
              }
            }
          },
          "#{NS_WSMAN_DMTF}:MaxEnvelopeSize" => session_opts[:max_envelope_size],
          "#{NS_ADDRESSING}:MessageID" => "uuid:#{SecureRandom.uuid.to_s.upcase}",
          "#{NS_WSMAN_MSFT}:SessionId" => "uuid:#{session_opts[:session_id]}",
          "#{NS_WSMAN_DMTF}:Locale/" => '',
          "#{NS_WSMAN_MSFT}:DataLocale/" => '',
          # "#{NS_WSMAN_DMTF}:OperationTimeout" =>
          #   Iso8601Duration.sec_to_dur(session_opts[:operation_timeout]),
          :attributes! => {
            "#{NS_WSMAN_DMTF}:MaxEnvelopeSize" => { 'mustUnderstand' => true },
            "#{NS_WSMAN_DMTF}:Locale/" => {
              'xml:lang' => session_opts[:locale], 'mustUnderstand' => false
            },
            "#{NS_WSMAN_MSFT}:DataLocale/" => {
              'xml:lang' => session_opts[:locale], 'mustUnderstand' => false
            },
            "#{NS_WSMAN_MSFT}:SessionId" => { 'mustUnderstand' => false }
          }
        }
      end

      # Helper methods for SOAP Headers

      def resource_uri_shell(shell_uri)
        {
          "#{NS_WSMAN_DMTF}:ResourceURI" => shell_uri, :attributes! => {
            "#{NS_WSMAN_DMTF}:ResourceURI" => {
              'mustUnderstand' => true
            }
          }
        }
      end

      def resource_uri_cmd
        resource_uri_shell(RESOURCE_URI_CMD)
      end

      def resource_uri_wmi(namespace = 'root/cimv2/*')
        {
          "#{NS_WSMAN_DMTF}:ResourceURI" =>
          "http://schemas.microsoft.com/wbem/wsman/1/wmi/#{namespace}",
          :attributes! => {
            "#{NS_WSMAN_DMTF}:ResourceURI" => {
              'mustUnderstand' => true
            }
          }
        }
      end

      def action_create
        {
          "#{NS_ADDRESSING}:Action" =>
          'http://schemas.xmlsoap.org/ws/2004/09/transfer/Create',
          :attributes! => {
            "#{NS_ADDRESSING}:Action" => {
              'mustUnderstand' => true
            }
          }
        }
      end

      def action_delete
        {
          "#{NS_ADDRESSING}:Action" =>
          'http://schemas.xmlsoap.org/ws/2004/09/transfer/Delete',
          :attributes! => {
            "#{NS_ADDRESSING}:Action" => {
              'mustUnderstand' => true
            }
          }
        }
      end

      def action_command
        {
          "#{NS_ADDRESSING}:Action" =>
          'http://schemas.microsoft.com/wbem/wsman/1/windows/shell/Command',
          :attributes! => {
            "#{NS_ADDRESSING}:Action" => {
              'mustUnderstand' => true
            }
          }
        }
      end

      def action_send
        {
          "#{NS_ADDRESSING}:Action" =>
          'http://schemas.microsoft.com/wbem/wsman/1/windows/shell/Send',
          :attributes! => {
            "#{NS_ADDRESSING}:Action" => {
              'mustUnderstand' => true
            }
          }
        }
      end

      def action_receive
        {
          "#{NS_ADDRESSING}:Action" =>
          'http://schemas.microsoft.com/wbem/wsman/1/windows/shell/Receive',
          :attributes! => {
            "#{NS_ADDRESSING}:Action" => {
              'mustUnderstand' => true
            }
          }
        }
      end

      def action_signal
        {
          "#{NS_ADDRESSING}:Action" =>
          'http://schemas.microsoft.com/wbem/wsman/1/windows/shell/Signal',
          :attributes! => {
            "#{NS_ADDRESSING}:Action" => {
              'mustUnderstand' => true
            }
          }
        }
      end

      def action_enumerate
        {
          "#{NS_ADDRESSING}:Action" =>
          'http://schemas.xmlsoap.org/ws/2004/09/enumeration/Enumerate',
          :attributes! => {
            "#{NS_ADDRESSING}:Action" => {
              'mustUnderstand' => true
            }
          }
        }
      end

      def selector_shell_id(shell_id)
        {
          "#{NS_WSMAN_DMTF}:SelectorSet" => {
            "#{NS_WSMAN_DMTF}:Selector" => shell_id, :attributes! => {
              "#{NS_WSMAN_DMTF}:Selector" => {
                'Name' => 'ShellId'
              }
            }
          }
        }
      end
    end
  end
end
